#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2020  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##
from waptserver.config import __version__

from waptcrypto import sha1_for_file
import sys
import os

try:
    wapt_root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
except:
    wapt_root_dir = 'c:/tranquilit/wapt'

import email.utils
import random
import stat
import time
import uuid
import urllib.parse

from waptserver.config import __version__
from waptserver.config import DEFAULT_CONFIG_FILE, load_config
from waptserver.utils import *
from waptserver.model import *
from waptenterprise.waptserver.wsus_common import *
from waptenterprise.waptserver.repositories import update_file_tree_of_files
from waptserver.tasks import *

waptwua_folder = conf['waptwua_folder']


@huey.task(retries=3, retry_delay=60, context=True)
def download_wsusscan(force=False, task=None):
    """Launch a task to update current wsus offline cab file
        download in a temporary well known file
        abort if the temporary file is present (means another download is in progress

    """
    with WaptDB():
        in_progress = WsusDownloadTasks.select(WsusDownloadTasks.id).where(
            (WsusDownloadTasks.kind == 'WSUSSCAN') &
            (~WsusDownloadTasks.status.in_(['OK', 'ERROR']))
        )

        if list(in_progress):
            logger.warning('Wsusscan, download already in progress, aborting.')
            return 'SKIPPED'

        cab_url = 'http://download.windowsupdate.com/microsoftupdate/v6/wsusscan/wsusscn2.cab'
        wsus_filename = os.path.join(waptwua_folder, 'wsusscn2.cab')
        tmp_filename = os.path.join(waptwua_folder, 'wsusscn2.cab.tmp_%s' % uuid.uuid1())

        wsusscan2_stats = WsusDownloadTasks(task_id=task.id, skipped=False)
        wsusscan2_stats.kind = 'WSUSSCAN'
        wsusscan2_stats.status = 'INIT'
        wsusscan2_stats.forced = force
        wsusscan2_stats.started_on = datetime.datetime.now()
        wsusscan2_stats.url = cab_url
        wsusscan2_stats.local_filename = tmp_filename
        wsusscan2_stats.save()  # be sure to get an id to not duplicate inserts

        http_proxy = get_http_proxies(conf)

        # should we remove the tmp filename when we get an error?
        cleanup_on_error = False

        if os.path.isfile(tmp_filename):
            os.unlink(tmp_filename)

        try:
            reply = requests.head(cab_url, proxies=http_proxy)
            last_modified = reply.headers['last-modified']
            wsusscan2_stats.file_date = httpdatetime2isodate(last_modified)

            new_cab_timestamp = httpdatetime2time(last_modified)
            if os.path.isfile(wsus_filename):
                current_cab_timestamp = fileutcmtime(wsus_filename)
            else:
                current_cab_timestamp = 0.0
            logger.info('download_wsuscan: current cab timestamp: %f, New cab timestamp: %f' % (current_cab_timestamp, new_cab_timestamp))

            if not os.path.isfile(wsus_filename) or (new_cab_timestamp > current_cab_timestamp) or force:
                logger.info('Downloading because of: file not found == %s, timestamps == %s, force == %s',
                            str(os.path.isfile(wsus_filename) == False), str(new_cab_timestamp > current_cab_timestamp), force)

                wsusscan2_stats.status = 'DOWNLOADING'
                wsusscan2_stats.target_size = None
                if 'content-length' in reply.headers:
                    wsusscan2_stats.target_size = int(reply.headers['content-length'])
                wsusscan2_stats.file_size = None
                wsusscan2_stats.save()
                start_time = time.time()

                def download_wsusscan_callback(downloaded_bytes, total_bytes, speed, url):
                    wsusscan2_stats.status = 'DOWNLOADING'
                    wsusscan2_stats.file_size = downloaded_bytes
                    wsusscan2_stats.target_size = total_bytes
                    if not wsusscan2_stats.started_on:
                        wsusscan2_stats.started_on = datetime.datetime.now()

                    if downloaded_bytes == 0:
                        wsusscan2_stats.msg = "Downloading %s (%.1f Mb)" % (url, int(total_bytes)/1024/1024)
                    elif downloaded_bytes >= total_bytes:
                        wsusscan2_stats.finished_on = datetime.datetime.now()
                        wsusscan2_stats.msg = "  -> download finished (%.0f Kb/s)" % (total_bytes / (1024.0*(time.time()+.001-start_time)))
                        wsusscan2_stats.status = 'FINISHED'
                    else:
                        wsusscan2_stats.msg = '%i / %i (%.0f%%) (%.0f KB/s)\r' % (downloaded_bytes, total_bytes, 100.0*downloaded_bytes/total_bytes, speed)
                    wsusscan2_stats.save()

                cleanup_on_error = True
                wget(cab_url, tmp_filename,
                     printhook=download_wsusscan_callback,
                     resume=True,
                     proxies=http_proxy
                     )

                file_stats = os.stat(tmp_filename)
                wsusscan2_stats.file_date = httpdatetime2isodate(email.utils.formatdate(file_stats[stat.ST_MTIME], usegmt=True))
                wsusscan2_stats.file_size = wsusscan2_stats.target_size = file_stats[stat.ST_SIZE]
                wsusscan2_stats.status = 'CHECKING'
                wsusscan2_stats.save()

                cablist = check_cab(tmp_filename).splitlines()
                wsusscan2_stats.cablist = cablist

                if os.path.isfile(wsus_filename):
                    os.unlink(wsus_filename)
                shutil.move(tmp_filename, wsus_filename)

                wsusscan2_stats.local_filename = wsus_filename
                wsusscan2_stats.status = 'OK'
                wsusscan2_stats.finished_on = datetime2isodate()
                wsusscan2_stats.save()

                sumwsus = sha1_for_file(wsus_filename)
                with open('%s.sha1' % wsus_filename, "w") as f:
                    f.write(sumwsus)

                if conf.get('remote_repo_support') or Hosts.select().where(Hosts.wapt_status.contains({'is_remote_repo':True})).count():
                    update_file_tree_of_files(username='TASK_DOWNLOAD_WSUS_SCAN')
                return 'OK'

            else:
                wsusscan2_stats.finished_on = datetime.datetime.now()
                wsusscan2_stats.msg = 'Skipped, already up to date'
                wsusscan2_stats.status = 'OK'
                wsusscan2_stats.skipped = True
                wsusscan2_stats.local_filename = wsus_filename
                file_stats = os.stat(wsusscan2_stats.local_filename)
                wsusscan2_stats.file_size = file_stats[stat.ST_SIZE]
                #wsusscan2_stats.finished_on = datetime.datetime.now()
                wsusscan2_stats.save()

                if not os.path.isfile('%s.sha1' % wsus_filename):
                    sumwsus = sha1_for_file(wsus_filename)
                    with open('%s.sha1' % wsus_filename, "w") as f:
                        f.write(sumwsus)

                return 'SKIPPED'

        except Exception as e:
            wsusscan2_stats.status = 'ERROR'
            wsusscan2_stats.error = ensure_unicode(e)
            if not wsusscan2_stats.started_on:
                wsusscan2_stats.finished_on = datetime.datetime.now()
            wsusscan2_stats.save()

            if cleanup_on_error and os.path.isfile(tmp_filename):
                os.unlink(tmp_filename)

            logger.error("Error in download_wsusscan: %s", str(e))
            logger.error('Trace:\n%s', traceback.format_exc())
            return 'ERROR'


def get_update_local_path(url):
    """Given a URK, return the local path of file in waptwua folder"""
    fileparts = urllib.parse.urlparse(url).path.split('/')
    target = os.path.join(waptwua_folder, fileparts[-1])
    return target

# Between 2h00 and 2h59


@huey.periodic_task(crontab(hour='3', minute=str(30 + random.randrange(-30, +30))))
def download_wsusscan_crontab():
    return download_wsusscan()


@huey.task(retries=3, retry_delay=60, context=True)
def download_windows_update_task(url, force=False, task=None):
    """Download a payload for windows updates given its URL

    The URL document part can contains the sha1 of the file.
    In this case, it is checked at the end of the download.

    Args:
        url
        force

    """
    with WaptDB():
        target = get_update_local_path(url)
        logger.info(r'Downloading %s to %s' % (url, target))

        # check sha1 sum if possible... return None if no sha1 in filename
        expected_sha1 = sha1_from_microsoft_filename(target)

        http_proxy = get_http_proxies(conf)

        download_task = WsusDownloadTasks(task_id=task.id, skipped=False)
        download_task.kind = 'WINUPDATE'
        download_task.url = url
        download_task.status = 'INIT'
        download_task.local_filename = target
        download_task.started_on = datetime.datetime.now()
        download_task.save()

        start_time = time.time()

        def progress_hook(received, total, speed, url):
            download_task.status = 'DOWNLOADING'
            download_task.file_size = received
            download_task.target_size = total
            if not download_task.started_on:
                download_task.started_on = datetime.datetime.now()

            if received == 0:
                download_task.msg = "Downloading %s (%.1f Mb)" % (url, int(total)/1024/1024)
            elif received >= total:
                download_task.finished_on = datetime.datetime.now()
                download_task.msg = "  -> download finished (%.0f Kb/s)" % (total / (1024.0*(time.time()+.001-start_time)))
                download_task.status = 'FINISHED'
            else:
                download_task.msg = '%i / %i (%.0f%%) (%.0f KB/s)\r' % (received, total, 100.0*received/total, speed)
            download_task.save()
        try:
            wget(url, target, resume=not force, sha1=expected_sha1, proxies=http_proxy, printhook=progress_hook)

            downloaded = True
            download_task.local_filename = target
            download_task.status = 'OK'
            download_task.finished_on = datetime.datetime.now()
            download_task.save()
        except Exception as e:
            download_task.status = 'ERROR'
            download_task.error = ("%s" % e)[:1200]
            download_task.finished_on = datetime.datetime.now()
            download_task.save()

        target_updates = WsusUpdates.filter(WsusUpdates.download_urls.contains(url))
        for update in target_updates:
            for update_url in update.download_urls:
                if not os.path.isfile(get_update_local_path(update_url)):
                    downloaded = False
                    break

            if downloaded:
                update.downloaded_on = datetime2isodate()
                update.save()
            elif not downloaded and update.downloaded_on:
                update.downloaded_on = None
                update.save()
        if conf.get('remote_repo_support') or Hosts.select().where(Hosts.wapt_status.contains({'is_remote_repo':True})).count():
            update_file_tree_of_files(username='TASK_DOWNLOAD_WINDOWS_UPDATE')

        logger.info(r'target file %s correctly downloaded' % target)
        return target


@huey.periodic_task(crontab(minute='*/30'))
def download_missing_cabs():
    """Scans table of Windows updates as reported by hosts.
    If teh Win yupdate shoul dbe installed, but is not yet downloaded, a
    task is created to download the missing files


    """
    # pylint: disable=no-member
    with WaptDB():
        updates = (HostWsus.select(WsusUpdates.update_id, WsusUpdates.download_urls)
                   .join(WsusUpdates).distinct()
                   .where((HostWsus.installed == False) & (HostWsus.allowed == True)))
        result = []
        for update in updates:
            for url in update.update_id.download_urls:
                if authorized_url_wsus(url):
                    local_path = get_update_local_path(url)
                    if not os.path.isfile(local_path) and not url in result:
                        result.append(url)
        for url in result:
            if not authorized_url_wsus(url):
                logger.info('Url %s not authorized for download' % url)
                continue
            download_windows_update_task(url)
            logger.info('Queued download of %s cabs' % len(result))
    return result


@huey.lock_task('cleanup-stalled-downloads')
def cleanup_stalled_downloads():
    with WaptDB():
        return (WsusDownloadTasks.update(
            status='ERROR',
            finished_on=datetime.datetime.now(),
            error='Killed, stalled download',
        )
            .where(
                (~WsusDownloadTasks.status.in_(['OK', 'ERROR'])) &
                (~WsusDownloadTasks.task_id.in_([t.task_id for t in huey.pending()]) &
                 (WsusDownloadTasks.started_on < datetime.datetime.now() - datetime.timedelta(hours=4))
                 )
        ).execute()
        )
